#include <windows.h>

#include <assert.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "serial_win.h"
#include "util.h"
#include "log.h"

#define SERIAL_BLOCKSIZE 0x1000

static HANDLE portfd = INVALID_HANDLE_VALUE;

//prototypes declaration
static int xread(HANDLE fd, void *_buf, int count);
//static int xaread(HANDLE fd, void *_buf, int count);
static int xwrite(HANDLE fd, const void *_buf, int count);
static int xawrite(HANDLE fd, const void *_buf, int count);

static void print_error_win(void){
	LPVOID lpMsgBuf;
	
	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // ftHg
		(LPTSTR) &lpMsgBuf,
		0,
		NULL 
		);
	
	message(lpMsgBuf);
	messagebox(lpMsgBuf);
	
	LocalFree( lpMsgBuf );
}

/* open the serial port */
int serial_open(const char *dev){
	COMMTIMEOUTS timeouts;
	DCB dcb;

	portfd = CreateFile(dev, GENERIC_READ|GENERIC_WRITE,
			0, NULL, OPEN_EXISTING, 0, NULL);
	if (portfd == INVALID_HANDLE_VALUE){
		print_error_win();
		return -1;
	}

	timeouts.ReadIntervalTimeout = 0;
	timeouts.ReadTotalTimeoutMultiplier = 10;
	timeouts.ReadTotalTimeoutConstant = 1000;
	timeouts.WriteTotalTimeoutMultiplier = 10;
	timeouts.WriteTotalTimeoutConstant = 1000;
	SetCommTimeouts (portfd, &timeouts);
	GetCommState (portfd, &dcb);
	dcb.BaudRate = 9600;
	dcb.fBinary = TRUE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	if (!SetCommState (portfd, &dcb)){
print_error_win();
		return -1;
	}
	return 0;
}

/* close serial port and restore settings */
int serial_close(void){
	int ret;
	if(portfd >= 0){
		ret = CloseHandle(portfd);
		portfd = INVALID_HANDLE_VALUE;
		if(ret == 0){
print_error_win();
			return -1;
		}
	}
	return 0;
}

/* switch baud rate */

int serial_baud(int baud){
	DCB dcb;
	GetCommState (portfd, &dcb);
	dcb.BaudRate = baud;
	dcb.fBinary = TRUE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	SetCommState (portfd, &dcb);
	return 0;
}

/* wait for a character on the serial port */
unsigned char get_char(void){
	unsigned char c;
	if(portfd < 0){
		message("Not open Serial\n");
		return -1;
	}
	xread(portfd, &c, 1);
	return c;
}

/* send a character on the serial port */
int put_char(unsigned char c){
	if(portfd < 0){
		message("Not oprn Serial\n");
		return -1;
	}
	xwrite(portfd, &c, 1);
	return 0;
}

/* wait for a word on the serial port, LSB first */
unsigned long get_word(void){
	unsigned w;
	
	w = get_char() & 0xff;
	w |= (get_char() & 0xff) << 8;
	w |= (get_char() & 0xff) << 16;
	w |= (get_char() & 0xff) << 24;
	return w;
}

/* send a word on the serial port, little-endian */
int put_word(unsigned long w){
	put_char((unsigned char)(w));
	put_char((unsigned char)(w >> 8));
	put_char((unsigned char)(w >> 16));
	put_char((unsigned char)(w >> 24));
	return 0;
}

int put_block(const char *buf, int size){
	assert(portfd >= 0);
	xawrite(portfd, buf, size);
	return 0;
}

/* ask the target to read a byte of memory */
unsigned char target_read_byte(unsigned long addr){
	put_char('g');
	put_word(addr);
	return get_char();
}

/* tell the target to write a byte of memory */
int target_write_byte(unsigned long addr, unsigned char data){
	put_char('s');
	put_word(addr);
	put_char(data);
	return 0;
}

/* ask the target to read a word of memory */
unsigned long target_read_word(unsigned long addr){
	put_char('r');
	put_word(addr);
	return get_word();
}

/* tell the target to write a word of memory */
int target_write_word(unsigned long addr, unsigned long data){
	put_char('w');
	put_word(addr);
	put_word(data);
	return 0;
}

/* tell the target to write a block of memory */
int target_write_block(unsigned long addr, const char *buf,
			       int size, int progress)
{
	char str[256];
	if(portfd < 0){
		message("Not open Serial\n");
		return -1;
	}

	while (size > 0) {
		char checksum = 0;
		int step = min(size, SERIAL_BLOCKSIZE);
		
		put_char('W');
		put_word(addr);
		put_word(step);
		xawrite(portfd, buf, step);
		addr += step;
		size -= step;
		progress += step;
		while (step-- > 0) {
			checksum += *buf++;
		}
		if (checksum != (char)get_char()) {
			message("\nSerial checksum error\n");
			messagebox("\nSerial checksum error\n");
			return -1;
		}
		
		sprintf(str, "0x%08x\r", progress);
		message(str);
		fflush(NULL);
	}
	return 0;
}


static int xread(HANDLE fd, void *_buf, int count){
	char *buf = (char *)_buf;
	int ret;
	unsigned long readlen = 0;

	ret = ReadFile(fd, buf, count, &readlen, NULL);
	if(ret == 0 || readlen == 0){
		return -1;
	}
	return readlen;
}

#if 0
static int xaread(HANDLE fd, void *_buf, int count){
	char *buf = (char *)_buf;
	int nread, remain;
	for (remain = count; remain; remain -= nread)
		buf += (nread = xread(fd, buf, remain));
	return count;
}
#endif

static int xwrite(HANDLE fd, const void *_buf, int count){
	char *buf = (char *)_buf;
	int ret;
	unsigned long writelen = 0;

	ret = WriteFile(fd, buf, count, &writelen, NULL);
	if(ret == 0 || writelen == 0){
		return -1;
	}
	return writelen;

}

/* write entire buffer or die */
static int xawrite(HANDLE fd, const void *_buf, int count){
	char *buf = (char *)_buf;
	int nwritten, remain;
	for (remain = count; remain; remain -= nwritten)
		buf += (nwritten = xwrite(fd, buf, remain));
	return count;
}

#if !defined(u32)
typedef unsigned int u32;
#endif
#if !defined(u16)
typedef unsigned short u16;
#endif
#if !defined(u8)
typedef unsigned char u8;
#endif

int 
read_serial(u8 *buf, u32 size, u32 timeout_ms)
{
	COMMTIMEOUTS timeouts, timeouts_old;
	int ret;
	
	GetCommTimeouts(portfd, &timeouts_old);

	timeouts.ReadIntervalTimeout = 0;
	timeouts.ReadTotalTimeoutMultiplier = 10;
	timeouts.ReadTotalTimeoutConstant = 100;
	timeouts.WriteTotalTimeoutMultiplier = 10;
	timeouts.WriteTotalTimeoutConstant = 100;
	SetCommTimeouts(portfd, &timeouts);

	ret = xread(portfd, buf, size);

	SetCommTimeouts(portfd, &timeouts_old);

	return ret;
}
int 
write_serial(u8 *buf, u32 size, u32 interval_us)
{
	u32 i;
	if (interval_us > 1000000)
		interval_us = 1000000;

	interval_us = (interval_us ? interval_us + 1000 : 1) / 1000;

	for (i=0; i<size; i++) {
		xwrite(portfd, &buf[i], 1);

		if (interval_us) {
			SleepEx(interval_us, TRUE);
		}
	}
	return 0;
}
int 
write_serial_byte(u8 data, u32 interval_us)
{
	return write_serial(&data, 1, interval_us);
}